from ctypes import c_int
from ctypes import c_uint
from ctypes import c_uint32
from ctypes import c_void_p
from ctypes import c_char_p
from ctypes import c_long
from ctypes import c_ulong
from ctypes import c_short
from ctypes import byref
from ctypes import memset
from ctypes import sizeof
from ctypes import Structure
from ctypes import POINTER
from ctypes import pointer
from ctypes import CDLL
from ctypes import get_errno


""" Wrapper for ADC C library using ctypes
All ctypes functions have exactly the same names as the ones in the
C library. To each function corresponds python function that hides
C specific operations from the user. The names of these functions are
the sames as the ctypes functions without ADC prefix."""

class PyAdcConf(Structure):
    __ADC_CONF_LEN = 64          # fixme what to do about it
    ADC_CONF_TYPE_BRD = 0
    ADC_CONF_TYPE_CHN = 1
    ADC_CONF_TYPE_ACQ = 2
    ADC_CONF_TYPE_CUS = 3
    ADC_CONF_TYPE_TRG_EXT = 4
    ADC_CONF_TYPE_TRG_THR = 5
    ADC_CONF_TYPE_TRG_TIM = 6
    __ADC_CONF_TYPE_LAST_INDEX = 7

    # BOARD configuration
    ADC_CONF_BRD_STATUS = 0
    ADC_CONF_BRD_MAX_FREQ_HZ = 1
    ADC_CONF_BRD_MIN_FREQ_HZ = 2
    ADC_CONF_BRD_STATE_MACHINE_STATUS = 3
    ADC_CONF_BRD_N_CHAN = 4
    ADC_CONF_BRD_N_TRG_EXT = 5
    ADC_CONF_BRD_N_TRG_THR = 6
    ADC_CONF_BRD_N_TRG_TIM = 7
    ADC_CONF_UTC_TIMING_BASE_S = 8
    ADC_CONF_UTC_TIMING_BASE_T = 9
    ADC_CONF_UTC_TIMING_BASE_B = 10
    __ADC_CONF_BRD_ATTRIBUTE_LAST_INDEX = 11

    # CHANNEL configuration
    ADC_CONF_CHN_RANGE = 0
    ADC_CONF_CHN_TERMINATION = 1
    ADC_CONF_CHN_OFFSET = 2
    ADC_CONF_CHN_SATURATION = 3
    ADC_CONF_CHN_GAIN = 4
    __ADC_CONF_CHN_ATTRIBUTE_LAST_INDEX = 5

    # TRIGGER EXTERNAL configuration
    ADC_CONF_TRG_EXT_ENABLE = 0
    ADC_CONF_TRG_EXT_POLARITY = 1
    ADC_CONF_TRG_EXT_DELAY = 2
    __ADC_CONF_TRG_EXT_LAST_INDEX = 3

    # TRIGGER THRESHOLD configuration
    ADC_CONF_TRG_THR_ENABLE = 0
    ADC_CONF_TRG_THR_POLARITY = 1
    ADC_CONF_TRG_THR_DELAY = 2
    ADC_CONF_TRG_THR_THRESHOLD = 3
    ADC_CONF_TRG_THR_HYSTERESIS = 4
    __ADC_CONF_TRG_THR_LAST_INDEX = 5

    # TRIGGER TIMER configuration
    ADC_CONF_TRG_TIM_ENABLE = 0
    __ADC_CONF_TRG_TIM_LAST_INDEX = 1

    # ACQUISITION configuration
    ADC_CONF_ACQ_N_SHOTS = 0
    ADC_CONF_ACQ_POST_SAMP = 1
    ADC_CONF_ACQ_PRE_SAMP = 2
    ADC_CONF_ACQ_UNDERSAMPLE = 3
    ADC_CONF_ACQ_FREQ_HZ = 4
    ADC_CONF_ACQ_N_BITS = 5
    __ADC_CONF_ACQ_ATTRIBUTE_LAST_INDEX = 6

    _fields_ = [("type", c_int),
                ("dev_type", c_uint),
                ("route_to", c_uint),
                ("flags", c_uint),
                ("mask", c_ulong),
                ("value", c_uint * __ADC_CONF_LEN)]

    def __init__(self, config_type, route_to):
        memset(byref(self), 0, sizeof(PyAdcConf))
        self.type = config_type
        self.route_to = route_to

    def mask_set(self, index):
        libadc.adc_set_conf_mask(byref(self), index)

    def mask_set_from_device(self, adc):
        libadc.adc_set_conf_mask_all(byref(self), adc)

    def value_set(self, index, value):
        libadc.adc_set_conf(byref(self), index, value)

    def value_get(self, index):
        value = c_uint32()
        libadc.adc_get_conf(byref(self), index, pointer(value))
        return value.value


class adc_buf(Structure):
    _fields_ = [("data", POINTER(c_short)),
                ("metadata", c_void_p),
                ("samplesize", c_int),
                ("nsamples", c_int),
                ("dev", c_void_p),
                ("mapaddr", c_void_p),
                ("maplen", c_uint),
                ("flags", c_uint)]


class timeval(Structure):
    _fields_ = [("sec", c_long),
                ("usec", c_long)]


class adc_timestamp(Structure):
    _fields_ = [("secs", c_ulong),
                ("ticks", c_ulong),
                ("bins", c_ulong)]

def libadc_create():
    def error_check_int(ret, func, args):
        """Generic error checker for functions returning 0 as success
        and -1 as error"""
        if ret < 0:
            raise OSError(get_errno(),
                          adc_strerror(get_errno()), "")
        else:
            return ret

    def error_check_pointer(ret, func, args):
        """Generic error handler for functions returning pointers"""
        if ret is None:
            raise OSError(get_errno(),
                          adc_strerror(get_errno()), "")
        else:
            return ret

    libadc = CDLL("libadc.so", use_errno=True)

    libadc.adc_strerror.argtypes = [c_int]
    libadc.adc_strerror.restype = c_char_p

    libadc.adc_print_version.restype = None
    libadc.adc_print_version.argtypes = None

    libadc.adc_init.restype = c_int
    libadc.adc_init.argtypes = None
    libadc.adc_init.errcheck = error_check_int

    libadc.adc_exit.restype = None
    libadc.adc_exit.argtypes = None

    libadc.adc_set_conf.restype = None
    libadc.adc_set_conf.argtypes = [POINTER(PyAdcConf), c_uint, c_uint]

    libadc.adc_set_conf_mask.restype = None
    libadc.adc_set_conf_mask.argtypes = [POINTER(PyAdcConf), c_uint]

    libadc.adc_get_conf.restype = c_int
    libadc.adc_get_conf.argtypes = [POINTER(PyAdcConf), c_uint,
                                    POINTER(c_uint32)]
    libadc.adc_get_conf.errcheck = error_check_int

    libadc.adc_get_driver_type.restype = c_char_p
    libadc.adc_get_driver_type.argtypes = [c_void_p]

    libadc.adc_open.restype = c_void_p
    libadc.adc_open.argtypes = [c_char_p, c_uint, c_ulong, c_uint, c_ulong]
    libadc.adc_open.errcheck = error_check_pointer

    libadc.adc_open_by_lun.restype = c_void_p
    libadc.adc_open_by_lun.argtypes = [c_char_p, c_int, c_ulong, c_uint,
                                     c_ulong]
    libadc.adc_open_by_lun.errcheck = error_check_pointer

    libadc.adc_request_buffer.restype = POINTER(adc_buf)
    libadc.adc_request_buffer.argtypes = [c_void_p, c_int, c_void_p, c_uint]
    libadc.adc_request_buffer.errcheck = error_check_pointer

    libadc.adc_apply_config.restype = c_int
    libadc.adc_apply_config.argtypes = [c_void_p, c_uint, c_void_p]
    libadc.adc_apply_config.errcheck = error_check_int

    libadc.adc_acq_start.restype = c_int
    libadc.adc_acq_start.argtypes = [c_void_p, c_uint, c_void_p]
    libadc.adc_acq_start.errcheck = error_check_int

    libadc.adc_acq_poll.restype = c_int
    libadc.adc_acq_poll.argtypes = [c_void_p, c_uint, c_void_p]
    libadc.adc_acq_poll.errcheck = error_check_int

    libadc.adc_fill_buffer.restype = c_int
    libadc.adc_fill_buffer.argtypes = [c_void_p, c_void_p, c_uint, c_void_p]
    libadc.adc_fill_buffer.errcheck = error_check_int

    libadc.adc_acq_stop.restype = c_int
    libadc.adc_acq_stop.argtypes = [c_void_p, c_uint]
    libadc.adc_acq_stop.errcheck = error_check_int

    libadc.adc_tstamp_buffer.restype = POINTER(adc_timestamp)
    libadc.adc_tstamp_buffer.argtypes = [c_void_p, c_void_p]
    libadc.adc_tstamp_buffer.errcheck = error_check_pointer

    libadc.adc_release_buffer.restype = c_int
    libadc.adc_release_buffer.argtypes = [c_void_p, c_void_p, c_void_p]
    libadc.adc_release_buffer.errcheck = error_check_int

    libadc.adc_close.restype = c_int
    libadc.adc_close.argtypes = [c_void_p]
    libadc.adc_close.errcheck = error_check_int

    libadc.adc_trigger_fire.restype = c_int
    libadc.adc_trigger_fire.argtypes = [c_void_p]
    libadc.adc_trigger_fire.errcheck = error_check_int

    libadc.adc_has_trigger_fire.restype = c_int
    libadc.adc_has_trigger_fire.argtypes = [c_void_p]

    """this function doesn't have body
libadc.adc_reset_conf.restype = c_int
libadc.adc_reset_conf.argtypes = [c_void_p, c_uint, c_void_p]
libadc.adc_reset_conf.errcheck = error_check_int"""

    libadc.adc_retrieve_config.restype = c_int
    libadc.adc_retrieve_config.argtypes = [c_void_p, c_void_p]
    libadc.adc_retrieve_config.errcheck = error_check_int

    libadc.adc_get_capabilities.restype = c_ulong
    libadc.adc_get_capabilities.argtypes = [c_void_p, c_int]
    libadc.adc_get_capabilities.errcheck = error_check_int

    libadc.adc_get_param.restype = c_int
    libadc.adc_get_param.argtypes = [c_void_p, c_char_p, c_char_p, c_void_p]
    libadc.adc_get_param.errcheck = error_check_int

    libadc.adc_set_param.restype = c_int
    libadc.adc_set_param.argtypes = [c_void_p, c_char_p, c_char_p, c_void_p]
    libadc.adc_set_param.errcheck = error_check_int

    libadc.adc_set_conf_mask_all.restype = None
    libadc.adc_set_conf_mask_all.argtypes = [c_void_p, c_void_p]

    return libadc

libadc = libadc_create()

def adc_strerror(err):
    return libadc.adc_strerror(err)

class ADC_Generic():
    ADC_F_USERMASK = 0xffff0000
    ADC_F_FLUSH = 0x00010000
    ADC_F_VERBOSE = 0x00020000
    ADC_F_FIXUP = 0x00400000

    ADC_TRG_POL_POS = 0
    ADC_TRG_POL_NEG = 1

    def print_version(self):
        self.adc_print_version()

    def __init__(self, devid, board=None):
        self.__ADC_CONF_LEN = 64
        """number of allocated items in each structure"""

        if not hasattr(self, 'BOARD_NAME'):
            raise Exception("Missing board name")

        self.__init()
        self.__adc_ptr = self.__open(devid, 0, 0, self.ADC_F_FLUSH)

    def __del__(self):
        self.__close()
        self.__exit()

    def __init(self):
        libadc.adc_init()

    def __exit(self):
        libadc.adc_exit()

    def get_driver_type(self):
        return libadc.adc_get_driver_type(self.__adc_ptr)

    def __open(self, dev_id, totalsamples, nbuffer, flags):
        return libadc.adc_open(bytes(self.BOARD_NAME, "ascii"), dev_id,
                             totalsamples, nbuffer, flags)

    def __open_by_lun(self, name, lun, totalsamples, nbuffer, flags):
        return libadc.adc_open_by_lun(bytes(self.BOARD_NAME, "ascii"), lun,
                                    totalsamples, nbuffer, flags)

    def request_buffer(self, nsamples, alloc, flags):
        return libadc.adc_request_buffer(self.__adc_ptr, nsamples, alloc, flags)

    def acq_start(self, flags, timeout):
        libadc.adc_acq_start(self.__adc_ptr, flags, timeout)

    def acq_poll(self, flags, timeout):
        libadc.adc_acq_poll(self.__adc_ptr, flags, timeout)

    def fill_buffer(self, buf, flags, timeout):
        libadc.adc_fill_buffer(self.__adc_ptr, buf, flags, timeout)

    def acq_stop(self, flags):
        libadc.adc_acq_stop(self.__adc_ptr, flags)

    def tstamp_buffer(self, buf, ts):
        libadc.adc_tstamp_buffer(buf, ts)

    def release_buffer(self, buf, free):
        libadc.adc_release_buffer(self.__adc_ptr, buf, free)

    def __close(self):
        if hasattr(self, "__adc_ptr"):
            libadc.adc_close(self.__adc_ptr)
            del self.__adc_ptr

    def trigger_fire(self):
        libadc.adc_trigger_fire(self.__adc_ptr)

    def has_trigger_fire(self):
        return libadc.adc_has_trigger_fire(self.__adc_ptr)

    def apply_config(self, config, flags):
        libadc.adc_apply_config(self.__adc_ptr, flags, byref(config))

    def retrieve_config(self, config):
        libadc.adc_retrieve_config(self.__adc_ptr, byref(config))

    def get_capabilities(self, type):
        libadc.adc_get_capabilities(self.__adc_ptr, type)

    def get_param(self, name, sptr, iptr):
        libadc.adc_get_param(self.__adc_ptr, name, sptr, iptr)

    def set_param(self, name, sptr, iptr):
        libadc.adc_set_param(self.__adc_ptr, name, sptr, iptr)
